package main

import (
	"bytes"
	"encoding/json"
	"fmt"
	"log"
	"net/http"
	"os"
	"strconv"
	"strings"
	"time"

	"git.dvdrw.dev/nsmarter/scraper/scraper"
)

var Log = log.Default()

const defaultChunkSize = 5
const defaultLimitQuerySize = 0

var stationDbPath string
var innerCityOnly bool = false
var chunkSize = 5
var apiEndpoint string = "https://online.nsmart.rs/publicapi/v1/announcement/announcement.php"
var apiKey string
var limitQuerySize = 0
var timeoutPeriod = 180
var fillerUrl string = "localhost"

func parseEnvars() {
	for _, e := range os.Environ() {
		pair := strings.SplitN(e, "=", 2)
		switch pair[0] {
		case "STATIONS_DB_PATH":
			stationDbPath = pair[1]
		case "INNER_CITY_ONLY":
			innerCityOnly = pair[1] == "1"
		case "CHUNK_SIZE":
			var err error
			chunkSize, err = strconv.Atoi(pair[1])
			if err != nil {
				Log.Printf("WARN: Invalid value for CHUNK_SIZE. Falling back to default value (%v)\n", defaultChunkSize)
				chunkSize = defaultChunkSize
			}
		case "TIMEOUT_PERIOD":
			var err error
			timeoutPeriod, err = strconv.Atoi(pair[1])
			if err != nil {
				Log.Printf("WARN: Invalid value for TIMEOUT_PERIOD. Falling back to default value (%v)\n", 180)
				timeoutPeriod = 180
			}
		case "LIMIT_QUERY_SIZE":
			var err error
			limitQuerySize, err = strconv.Atoi(pair[1])
			if err != nil {
				Log.Printf("WARN: Invalid value for LIMIT_QUERY_SIZE. Falling back to default value (%v)\n", defaultLimitQuerySize)
				limitQuerySize = defaultLimitQuerySize
			}
		case "API_ENDPOINT":
			apiEndpoint = pair[1]
		case "API_KEY":
			apiKey = pair[1]
		case "FILLER_URL":
			fillerUrl = pair[1]
		}
	}
}

func main() {
	// Parse out config from environment variables
	parseEnvars()

	if stationDbPath == "" {
		log.Fatal("Environment variable STATIONS_DB_PATH empty!")
	}

	stations, err := readDb()
	if err != nil {
		log.Fatal(err)
	}

	Log.Printf("Finished importing stations! (Total: %v)", len(stations))

	// Split stations so as to be scraped in chunks
	var stationChunks [][]scraper.Station
	for i := 0; i < len(stations); i += chunkSize {
		end := i + chunkSize

		if end > len(stations) {
			end = len(stations)
		}

		stationChunks = append(stationChunks, stations[i:end])
	}

	results := make(chan []scraper.ScrapeResult, 200)
	go func() {
		for _, chunk := range stationChunks {
			go scraper.ScheduleScrape(chunk,
				results,
				scraper.ApiConfig{Endpoint: apiEndpoint,
					Key:     apiKey,
					Timeout: int64(timeoutPeriod)})
			// Fan out scrapes over time so as to avoid upstream rate limits
			time.Sleep(time.Millisecond * 100)
		}
	}()

	for r := range results {
		json, err := json.Marshal(r)
		if err != nil {
			fmt.Print("Couldn't serialise struct to JSON: ", err)
		}

		request, err := http.NewRequest("POST", fillerUrl+"/v1/submit", bytes.NewBuffer(json))
		request.Header.Set("Content-Type", "application/json; charset=UTF-8")

		go (func(req *http.Request) {
			res, err := http.DefaultClient.Do(req)
			if err != nil {
				fmt.Print("Error while sending data to filler: ", err)
			} else if res.StatusCode != 200 {
				fmt.Print("Non-200 while sending data to filler!")
			}
		})(request)
	}
}
